using System;
using System.Collections.Generic;
using System.Windows.Forms;
using Irrlicht;
using Irrlicht.Core;
using Irrlicht.Scene;
using Irrlicht.Video;

/*
 * 
 * TUTORIAL - Moteur "tile engine" trs trs basique avec Irrlicht.
 * Ecrit par Ambroise Garel (agar@cafedefaune.com) pour www.irrlicht.fr
 * 
 * Le but de ce tutorial est de dmonter au lecteur qu'il est extrmement simple de raliser un petit
 * moteur utlisant des "tiles" avec Irrlicht.
 * 
 * Cette version est trs basique, mais donne les outils principaux (reprage de l'endroit sur lequel le joueur
 * clique afin de dplacer son perso, etc...)
 * 
 * Ce moteur peut servir de base  la ralisation d'un Diablo-like, d'un "Zelda", d'un jeu de stratgie, etc...
 * 
 * Merci de bien vouloir me tenir au courant si vous avez de ce tutorial autre chose qu'un usage priv
 * (upload sur votre site, traduction, etc...). Pas pour vous crier dessus, juste pour tre au courant. Merci!
 * 
 * NB: Je sais que tous ces objets statiques ne sont pas beaux du tout, mais c'est un tutorial, pas un vrai
 * programme. J'ai cherch  faire au plus lisible et au plus simple.
 * 
 * */

namespace TileTutorial
{
    public class TileEngine
    {
        // La taille de la carte
        const int MAPSIZE_X = 32;
        const int MAPSIZE_Y = 32;

        // La taille de la fentre d'affichage
        const int WINDOWSIZE_X = 800;
        const int WINDOWSIZE_Y = 600;

        // Le device Irrlicht et ses petits.
        public static IrrlichtDevice dev;
        public static ISceneManager scm;
        public static IVideoDriver drv;

        // Le receveur d'vnement, une classe nomme MyEventReceiver (voir plus bas)
        public static MyEventReciever evnt;

        // La camera
        public static ICameraSceneNode cam;

        // Les tiles
        public static ISceneNode[,] FloorTile = new ISceneNode[MAPSIZE_X, MAPSIZE_Y];

        // Le joueur
        public static IAnimatedMeshSceneNode Player;

        // La position o le joueur doit aller.
        public static Vector3D playertargetpos;

        // Une source de lumire
        public static ILightSceneNode light;

        // Tant que Running est "true", le programme tourne.
        public static bool Running;

        // Fonction Main()
        // Tout commence ici.
        [STAThread]
        static void Main()
        {
            // On demande  l'utilisateur s'il souhaite utiliser OpenGL ou DirectX
            DialogResult DR;
            DR = MessageBox.Show("Souhaitez-vous utiliser OpenGL ? Si vous rpondez non, Direct3D 9.0c sera utilis.", "OpenGL ou non ?", MessageBoxButtons.YesNo, MessageBoxIcon.Question);

            DriverType DT;
            if (DR == DialogResult.Yes)
                DT = DriverType.OPENGL;
            else
                DT = DriverType.DIRECT3D9;

            // On demande  l'utilisateur s'il souhaite utiliser du normal mapping
            bool usenormals = false;
            DR = MessageBox.Show("Souhaitez-vous utiliser des normal maps ?", "Normal maps ou vieilles textures  l'ancienne ?", MessageBoxButtons.YesNo, MessageBoxIcon.Question);

            if (DR == DialogResult.Yes)
                usenormals = true;

            // Cration du device Irrlicht
            dev = new IrrlichtDevice(DT, new Dimension2D(WINDOWSIZE_X, WINDOWSIZE_Y), 32, false, false, false, false, new IntPtr());
            scm = dev.SceneManager;
            drv = dev.VideoDriver;

            dev.WindowCaption = "Pressez chap pour quitter";

            // Cration du receveur d'vnements
            evnt = new MyEventReciever();
            dev.EventReceiver = evnt;

            // On charge le mesh utilis pour le sol.
            IMesh TileMesh = scm.GetMesh(@"..\..\media\floor.x").GetMesh(0);
            // Il faut crer les tangeantes pour utiliser le normal mapping.
            if (usenormals) TileMesh = scm.MeshManipulator.CreateMeshWithTangents(TileMesh);

            // Et maintenant la petite texture qui va avec.
            ITexture TileTex = drv.GetTexture(@"..\..\media\floor.jpg");

            // Et la normal map.
            ITexture TileTexNormal = drv.GetTexture(@"..\..\media\floor_normal.jpg");

            // Cration du monde...

            // 1- D'abord, on ajoute les "tiles" du sol
            for (int i = 0; i < MAPSIZE_X; i++)
                for (int j = 0; j < MAPSIZE_Y; j++)
                {
                    // Trs important : les nodes des tiles ont une ID de 1! Cela servira lors de la dtection
                    // des clics.
                    FloorTile[i, j] = scm.AddMeshSceneNode(TileMesh, null, 1);
//                    FloorTile[i, j].SetMaterialFlag(MaterialFlag.LIGHTING, false);
                    FloorTile[i, j].SetMaterialTexture(0, TileTex);
                    FloorTile[i, j].Position = new Vector3D(i * 5, 0, j * 5);

                    if (usenormals)
                    {
                        FloorTile[i, j].SetMaterialTexture(1, TileTexNormal);
                        FloorTile[i, j].SetMaterialType(MaterialType.NORMAL_MAP_SOLID);
                    }
                }

            // Et voil votre incarnation : syndney.md2
            Player = scm.AddAnimatedMeshSceneNode(scm.GetMesh(@"..\..\media\sydney.md2"), null, 0);
            Player.SetMaterialTexture(0, drv.GetTexture(@"..\..\media\sydney.bmp"));
            Player.Position = new Vector3D(2.5f, 2.5f, 2.5f);
            // Syndey est un peu grosse, alors on va la rduire.
            Player.Scale = new Vector3D(0.15f, 0.15f, 0.15f);
            Player.SetMD2Animation(MD2AnimationType.STAND);

            // Pour le moment, le joueur ne va nulle part, donc sa position cible sera sa position actuelle.
            playertargetpos = Player.Position;

            // On ajoute la camra...
            cam = scm.AddCameraSceneNode(null, new Vector3D(0, 25, 25), new Vector3D(0, 0, 0), 0);

            // Et une petite source de lumire, histoire qu'on y voit quelque chose.
            light = scm.AddLightSceneNode(null, new Vector3D(25, 25, 25), new Colorf(1.0f, 0.95f, 0.9f), 100, 0);

            // C'est parti! Tant que running est "true", la boucle principale tourne.
            Running = true;

            // Et voici donc cette boucle de dessin principale.
            DrawLoop();

            // On est sorti de la boucle principale car la fentre n'est plus active.
            // Il ne reste plus...

            // 1- Qu' dtruire le device
            dev.CloseDevice();

            // 2- Et  faire un peu le mnage dans la mmoire.
            GC.Collect();
        }

        static void DrawLoop()
        {
            // Boucle tant que la fentre est active.
            while (Running)
            {
                // On dessine toutes ces petites choses.
                drv.BeginScene(true, true, new Color(0, 0, 0, 0));
                scm.DrawAll();
                drv.EndScene();

                // On laisse un peu respirer l'application.
                Application.DoEvents();

                // Si le joueur ne se trouve pas  sa position cible, il faut qu'il bouge...
                if (
                    (playertargetpos.X != Player.Position.X) ||
                    (playertargetpos.Z != Player.Position.Z)
                   )
                {
                    float speedx = 0.0f;
                    float speedz = 0.0f;

                    // On vrifie l'axe X
                    if (Player.Position.X < playertargetpos.X) speedx = 0.1f;
                    else if (Player.Position.X > playertargetpos.X) speedx = -0.1f;

                    // On vrifie l'axe Z
                    if (Player.Position.Z < playertargetpos.Z) speedz = 0.1f;
                    else if (Player.Position.Z > playertargetpos.Z) speedz = -0.1f;

                    // On fait bouger tout a...
                    Player.Position = new Vector3D(Player.Position.X + speedx, 2.5f, Player.Position.Z + speedz);

                    //Pour viter un effet de "sautillement", si le joueur est suffisemment prs de sa position
                    //cible, on l'y place directement.
                    if (Math.Abs(Player.Position.X - playertargetpos.X) < 0.15f)
                        Player.Position = new Vector3D(playertargetpos.X, 2.5f, Player.Position.Z);

                    if (Math.Abs(Player.Position.Z - playertargetpos.Z) < 0.15f)
                        Player.Position = new Vector3D(Player.Position.X, 2.5f, playertargetpos.Z);
                }

                // La camra suit le perso
                cam.Position = new Vector3D(Player.Position.X + 10, 15, Player.Position.Z + 10);
                cam.Target = Player.Position;

                // La lumire suit le joueur, puisqu'il a un "ring of light +1"
                // Je ne vous l'avait pas dit? Maintenant, c'est fait.
                light.Position = new Vector3D(Player.Position.X +10, 5, Player.Position.Z +10);
            }
        }
    }

    public class MyEventReciever : IEventReceiver
    {
        // Lecture des vnements Irrlicht (en l'occurence, les clics de souris)
        public bool OnEvent(Event e)
        {
            if (e.Key == KeyCode.KEY_ESCAPE)
                TileEngine.Running = false;

            // Si c'est pas la souris, poubelle. Il n'y a que la souris qui nous intresse.
            if (e.Type != EventType.MouseInput) return false;

            // Un clic gauche... Ah, ah! Ca veut dire que le joueur veut bouger.
            if (e.MouseInputType == MouseInputEvent.PressedDownLeft)
            {
                // On doit trouver sur quel tile le joueur a cliqu.
                // Ca se fait avec le SceneCollisionManager.
                // Notez que le paramtre "idbitmask" permet de ne prendre en compte que certains nodes.
                // (Ici, ceux avec l'ID 1, donc les tiles. Cliquer sur Sydney ne fait donc rien).
                ISceneNode ClickedNode =
                    TileEngine.scm.SceneCollisionManager.GetSceneNodeFromScreenCoordinatesBB(e.MousePos, 1, true);

                // Si le joueur n'a pas cliqu sur une tile, on oublie tout a...
                if (ClickedNode == null) return false;

                // Sinon, le joueur doit maintenant aller sur cette tile.
                // On ajoute 2,5 aux coordonnes X et Z pour que le joueur aille au centre de la tile, pas
                // dans un coin.
                TileEngine.playertargetpos =
                    new Vector3D(ClickedNode.Position.X + 2.5f, 0, ClickedNode.Position.Z + 2.5f);
            }

            return false;
        }
    }
}

/*
 * 
 * Merci d'avoir lu mon tutorial.
 * Si vous avez des questions ou des suggestions, vous pouvez m'crire  l'adresse agar@cafedefaune.com.
 * 
 * QUELQUES SUGGESTIONS POUR ALLER PLUS LOIN:
 * ==========================================
 * 
 * 1- Dj, il serait bien que le perso soit anim et se tourne dans la direction o il se dplace. Si le premier
 * point est trs simple  rgler, le deuxime ncessite de sortir un peu de trigonomtrie.
 * 
 * 2- La carte est petite par dfaut (32x32), donc a tourne pas trop mal. Maintenant, avec une carte plus grande,
 * les tiles trop nombreuses risquent de noyer Irrlicht (et la carte vido) sous une avalanche de "nodes". Il
 * faudra donc crire une petite fonction qui dtermine si telle ou telle tile est visible  un moment prcis et
 * affiche dynamiquement uniquement les tiles ncessaires.
 * 
 * 3- En utilisant les "id" des nodes, pourquoi ne pas ajouter des ennemis? Ensuite, quand on clique, le programme
 * devra vrifier si l'ID du node cliqu est 1 (le sol) ou 2, par exemple. Si c'est 2, alors lancer la squence
 * d'attaque... Avec cette technique, il est trs trs rapide de crer un "petit diablo".
 * 
 * 4- Ajouter des murs n'est pas trs compliqu non plus. Il suffit d'utiliser le modle "floor.x" et de le faire
 * pivoter de 90 degrs.
 * 
 * 5- Qui dit mur dit obstacle. Et qui dit obstacle dit pathfinding. Ca peut tre intressant  ajouter, mais
 * c'est un peu plus de travail.
 * 
 * 6- Et allez, si vous avez du sommeil  perdre... Vous nous crirez bien un petit gnrateur alatoire de
 * donjons, non ?
 * 
*/